#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>

// Lossage for TSK
#ifdef linux
#define LINUX2
#endif


#include "tsk_os.h"
#include "aux_tools.h"
#include "fs_tools.h"
#include "mm_tools.h"

#include "afflib.h"

#include <algorithm>
#include <cstdlib>
#include <vector>
#include <string>

/* Note: TSK has a broken MD5 implementation, but we can't define a better
 * one because of some namespace conflicts. So just use this:
 */

#define MD5_Update MD5Update
#define MD5_Init   MD5Init
#define MD5_Final  MD5Final



/* TTD: Modify Block Runs so it prints NN-NN if block is length 1, or change
 * output format to Block[+LEN]
 */

char *iwalk_version = "0.3";
int do_orphan=1;
int do_allocated=1;
char *outdir = "";

struct seg {
    int64 start;
    int64 len;
};

using namespace std;
typedef vector<struct seg> segs_;	// vector of blocks
segs_ segs;

int opt_debug    = 0;
int opt_md5      = 0;			// calculate the MD5 for every file?
int opt_save     = 0;

/* block_map (will) keeps track of all the blocks that have been touched
 * by the regular walking, so that it can be carved.
 */
unsigned char *block_map = 0;

/* Individual 'state' variables */
MD5_CTX md5;
FILE *outfile = 0;
string save_filename;

/* 
 * af_hexbuf:
 * Turn a binay string into a hex string, optionally with spaces.
 */

#define HEXBUF_NO_SPACES 0
#define HEXBUF_SPACE2    0x0001	// space every 2 characters
#define HEXBUF_SPACE4    0x0002	// space every 4 characters
#define HEXBUF_UPPERCASE 0x1000	// uppercase


const char *hexbuf(char *dst,int dst_len,const unsigned char *bin,int bytes,int flag)
{
    int charcount = 0;
    const char *start = dst;		// remember where the start of the string is
    const char *fmt = (flag & HEXBUF_UPPERCASE) ? "%02X" : "%02x";

    *dst = 0;				// begin with null termination
    while(bytes>0 && dst_len > 3){
	sprintf(dst,fmt,*bin); // convert the next byte
	dst += 2;
	bin += 1;
	dst_len -= 2;
	bytes--;
	charcount++;			// how many characters
	
	bool add_spaces = false;
	if(flag & HEXBUF_SPACE2) add_spaces = true;
	if((flag & HEXBUF_SPACE4) && charcount%2==0){
	    *dst++ = ' ';
	    *dst   = '\000';
	    dst_len -= 1;
	}
    }
    return start;			// return the start
}


/* FS_INFO defined in ~/sleuthkit-2.05/src/fstools/fs_tools.h
 * IMG_INFO defined in img_tools.h 
 * img_open - opens the image; calls do_dimage
 * do_dimage - analyizes the image
 *        mm_open - opens the volume; mm_part_walk - walk each volume;
 *                  calls mm_act for each
 * mm_act - analyizes each partition; calls do_vol for ones that we can analyze
 * do_vol - analyizes each volume
 *       fs_open - opens the file system; dent_walk 
 *               - walks the directory structure.
 *               - calls dent_act for each file
 * dent_act - prints file name; calls file_walk() to read each file; 
 *          - calls file_act for each sector
 * file_act - prints the block number
 */

/* 
 * dent_walk callback function -- print the address of each data unit
 */
uint8_t
file_act(TSK_FS_INFO * fs, DADDR_T addr, char *buf, size_t size, TSK_FS_BLOCK_FLAG_ENUM flags, void *ptr)
{
    if(opt_debug){
	printf("addr=%"PRIuDADDR" buf=%p size=%d\n",addr,buf,(int)size);
	if(opt_debug>1 && segs.size()==0){
	    fwrite(buf,size,1,stdout);
	    printf("\n");
	}
    }

    if(size==0) return TSK_WALK_CONT;	    // can't do much with this...
    if(opt_md5) MD5_Update(&md5,(unsigned char *)buf,size);
    if(opt_save) fwrite(buf,size,1,outfile);

    /* See if we can extend the last segment */
    int64 byteaddr = ((int64)addr)*fs->block_size;

    if(segs.size()>0){
	if(segs.back().start + segs.back().len == (int64)byteaddr){
	    segs.back().len += size;
	    return TSK_WALK_CONT;
	}
    }
    /* Need to add a new element to the list */
    struct seg newseg;
    newseg.start = byteaddr;
    newseg.len   = size;
    segs.push_back(newseg);
    return TSK_WALK_CONT;
}


void print_fsi(TSK_FS_INODE *fsi)
{
    printf("Size: %qd\n",(long long)fsi->size);
    printf("Status: ");
    if(fsi->flags & TSK_FS_INODE_FLAG_ALLOC) printf("ALLOC ");
    if(fsi->flags & TSK_FS_INODE_FLAG_UNALLOC) printf("UNALLOC ");
    if(fsi->flags & TSK_FS_INODE_FLAG_USED) printf("USED ");
    if(fsi->flags & TSK_FS_INODE_FLAG_UNUSED) printf("UNUSED ");
    if(fsi->flags & TSK_FS_INODE_FLAG_ORPHAN) printf("ORPHAN ");
    printf("\n");
    if(fsi->mtime) printf("mtime: (%d) %s",
				   (int)fsi->mtime,
				   ctime(&fsi->mtime));
    if(fsi->ctime) printf("ctime: (%d) %s",
				   (int)fsi->ctime,
				   ctime(&fsi->ctime));
    if(fsi->atime) printf("atime: (%d) %s",
				   (int)fsi->atime,
				   ctime(&fsi->atime));
}

/*
 * inode_act:
 * Called for each file in the file system
 * Callback action for the file system dent_walk
 *
 * Walk each file to print the addresses
 */
static uint8_t
inode_act(TSK_FS_INFO * fs, TSK_FS_INODE * fsi, void *ptr)
{
    /* If the name has corresponding metadata, then walk it */
    segs.clear();
    if (fsi) {
	if ((fsi->mode & TSK_FS_INODE_MODE_FMT)!=TSK_FS_INODE_MODE_REG){
	    /* Only do regular files */
	    return TSK_WALK_CONT;
	}

	TSK_FS_FILE_FLAG_ENUM myflags = TSK_FS_FILE_FLAG_NOID;
	
	//myflags |= TSK_FS_FLAG_FILE_SLACK;
	//myflags |= TSK_FS_FILE_FLAG_RECOVER;

	if(opt_md5){
	    /* We are going to get every single block, rather than
	     * just the metadata...
	     */
	    MD5_Init(&md5);
	}
	if(opt_save){
	    while(access(save_filename.c_str(),R_OK)==0){
		save_filename += ".1";
	    }
	    printf("open(%s)\n",save_filename.c_str());
	    outfile = fopen(save_filename.c_str(),"w");
	}
	if(!opt_md5 && !opt_save){
	    /* Just get the metadata */
	    //myflags |= TSK_FS_FILE_FLAG_AONLY;
	}
	
	if(ptr==NULL) printf("Orphan: YES\n");
	print_fsi(fsi);

	/* Now get the information on the individual
	 * sectors that make up the file.
	 */
	if (fs->file_walk(fs, fsi, 0, 0, myflags, file_act, NULL)) {
	    tsk_error_reset();
	}
	
	if(segs.size()>0){
	    printf("Byte runs: ");
	    for(segs_::const_iterator i = segs.begin();i!=segs.end();i++){
		printf("%qd:%qd ",i->start,i->len);
	    }
	    printf("\n");
	}
	
	printf("Fragments: %d\n",(int)segs.size());
	if(opt_md5){
	    unsigned char md5buf[16];
	    char buf[64];
	    MD5_Final(md5buf,&md5);
	    hexbuf(buf,sizeof(buf),md5buf,16,0);
	    printf("MD5: %s\n",buf);
	}
	if(opt_save){
	    fclose(outfile);
	    outfile = 0;
	}
	printf("\n");
	fflush(stdout);
    }
    return TSK_WALK_CONT;
}


/*
 * dent_acct:
 * Called for each file in the file system
 * Callback action for the file system dent_walk
 *
 * Walk each file to print the addresses
 */
static uint8_t
dent_act(TSK_FS_INFO * fs, TSK_FS_DENT * fs_dent, void *ptr)
{
    save_filename = "";
    if (fs_dent->fsi) {
	printf("Filename: %s%s\n", fs_dent->path, fs_dent->name);
	save_filename = string(fs_dent->path) + fs_dent->name;
	save_filename = string(outdir) + string("/") + fs_dent->name;
	return inode_act(fs,fs_dent->fsi,ptr);
    }
    return TSK_WALK_CONT;
}


/* 
 * Analyze the volume starting at byte offset 'start' in the image
 * file pointed to by img. 
 *
 * Return 0 on success and 1 on error 
 */
int do_vol(TSK_IMG_INFO *img, OFF_T start)
{
    TSK_FS_INFO *fs;
    int ret = 0;

    if (tsk_verbose) fprintf(stderr, "do_vol:  Start: %" PRIuDADDR "\n", start);

    /* Try it as a file system */
    if ((fs = tsk_fs_open(img, start, NULL)) == NULL) {
	tsk_error_print(stderr);

	/* @@@ We could do some file system carving at this point */
	return 1;
    }

    printf("Partition Offset: %"PRIuDADDR" (sector %"PRIuDADDR") ", fs->offset,fs->offset/512);
    printf("block_size=%d  ftype=%d  (%s) ",fs->block_size,fs->ftype,tsk_fs_get_type(fs->ftype));
    printf("block_count=%"PRIuDADDR"  ",fs->block_count);
    printf("first_block=%"PRIuDADDR"  ",fs->first_block);
    printf("last_block: %"PRIuDADDR"  ",fs->last_block);
    printf("\n");
    printf("\n");
    fflush(stdout);

    block_map = (unsigned char *)calloc(1,fs->last_block);

    /*
     * If we didn't want information on the files, we could just
     * return here with
     *
     * fs->close(fs); return 0;
     */
    

    if(do_orphan){
	time_t t = time(0);
	printf("# Performing orphan inode_walk from %qd to %qd at %s",
	       (long long)fs->first_inum,(long long)fs->last_inum,ctime(&t));

	/* Don't need other flags; it will fill in the rest */
	TSK_FS_INODE_FLAG_ENUM inode_walk_flags = TSK_FS_INODE_FLAG_ORPHAN; 
	if (fs->inode_walk(fs,fs->first_inum,fs->last_inum,inode_walk_flags, inode_act, NULL)){
	    tsk_error_print(stderr);
	    ret = 1;
	}
    }


    /* Walk the files, starting at the root directory */
    if(do_allocated){
	time_t t = time(0);
	printf("# Starting dent_walk at %s",ctime(&t));
	int dent_walk_flags = (TSK_FS_DENT_FLAG_ALLOC | TSK_FS_DENT_FLAG_UNALLOC |
			       TSK_FS_DENT_FLAG_RECURSE);
	if (fs->dent_walk(fs, fs->root_inum, (TSK_FS_DENT_FLAG_ENUM)dent_walk_flags, dent_act, &ret)) {
	    tsk_error_print(stderr);
	    ret = 1;
	}
    }
	

    fs->close(fs);
    free(block_map);
    return ret;
}

/*
 * Callback for the partition table walk.
 *
 * Analyze the file system for each partition.
 *
 */
uint8_t mm_act(TSK_MM_INFO * mm, PNUM_T pnum, TSK_MM_PART * part, int flag, void *ptr)
{
    // Currently, skip the "descriptive" entries
    if (part->type == TSK_MM_PART_TYPE_DESC){
	return TSK_WALK_CONT;
    }

    if (do_vol(mm->img_info, part->start * mm->block_size)) {
	return TSK_WALK_ERROR;
    }
    return TSK_WALK_CONT;
}



/* do_dimage:
 * Process the partition tables in the disk image pointed to by img.
 * File system analysis will be performed on each partition.
 *   mm_open() - opens partition
 *   mm_part_walk() - walks partition
 * If we can't open as a disk image, try to open as a raw file system
 *
 * returns 1 on error and 0 on success
 */
int do_dimage(TSK_IMG_INFO * img, char *desc)
{
    TSK_MM_INFO *mm;

    if ((mm = tsk_mm_open(img, 0, NULL)) == NULL) {
	if (tsk_verbose) fprintf(stderr, "Error determining volume system -- trying file systems\n");

	/* Check if we have a partition image */
	tsk_error_reset();
	if (do_vol(img, 0)) {
	    return 1;
	}
    }
    else {
	if (tsk_verbose) fprintf(stderr, "Volume system open, examining each\n");
	/* Walk the volumes and update the volume table */
	if (mm->part_walk(mm, mm->first_part, mm->last_part, 0, mm_act, 0)) {
	    return 1;
	}
    }
    return 0;
}


void usage()
{
    printf("usage: iwalk [options] iso-name\n");
    printf("Default behavior: Just print the file system statistics and exit.\n");
    printf("options:\n");
    printf("    -d = debug this program\n");
    printf("    -v = Enable SleuthKit verbose flag\n");
    printf("    -O = Just report orphan files\n");
    printf("    -F = Just report allocated files\n");
    printf("\nThese options slow down processing significantly:\n");
    printf("    -m = Also report MD5 for each file\n");
    printf("    -s <dir> = Save all recovered files to <dir>\n");
    printf("\n");
    exit(1);
}


int main(int argc, char *argv[])
{
    TSK_IMG_INFO *img;
    int ch;
    extern int optind;

    int t0 = time(0);

    while ((ch = getopt(argc, argv, "mdvOFs:")) > 0) {
	switch (ch) {
	case 'm':
	    opt_md5++;
	    break;
	case 'd':
	    opt_debug++;
	    break;
	case 'v':
	    tsk_verbose++;			// sleuthkit option
	    break;
	case 'F':
	    do_allocated=1;
	    do_orphan=0;
	    break;
	case 'O':
	    do_allocated=0;
	    do_orphan=1;
	    break;
	case 's':
	    outdir = optarg;
	    opt_save = 1;
	    break;
	case 'h':
	case '?':
	default:
	    fprintf(stderr, "Invalid argument: %s\n", argv[optind]);
	    usage();
	}
    }

    if (optind >= argc) usage();

    argc -= optind;
    argv += optind;

    /* First check that we have a valid file format */
    printf("# iwalk version %s\n",iwalk_version);
    const char *filename = argv[0];
    if ((img = tsk_img_open(NULL, argc, (const char **)argv)) == NULL) {
	tsk_error_print(stderr);
	printf("\n\n=EOF= (img_open)\n");
	exit(1);
    }
    printf("# %s\n",filename);


    printf("Imagefile: %s\n",filename);
    if (do_dimage(img, "my boring test comment")) {
	tsk_error_print(stderr);
	printf("\n\n=EOF= (do_dimage)\n");
	exit(1);
    }
    img->close(img);

#ifdef HAVE_GETRUSAGE
    /* Print usage information */
    struct rusage ru;
    memset(&ru,0,sizeof(ru));
    int t1 = time(0);
    printf("# clock: %d\n",t1-t0);
    if(getrusage(RUSAGE_SELF,&ru)==0){
	time_t t = time(0);
	printf("# u: %d  s: %d\n",(int)ru.ru_utime.tv_sec,(int)ru.ru_stime.tv_sec);
	printf("# maxrss: %ld reclaims: %ld faults: %ld   swaps: %ld\n",
	       ru.ru_maxrss,ru.ru_minflt,ru.ru_majflt,ru.ru_nswap);
	printf("# inputs: %d  outputs: %d\n",(int)ru.ru_inblock,(int)ru.ru_oublock);
	printf("# %s",ctime(&t));
    }
#endif
    printf("\n\n=EOF=\n");
    exit(0);
}

